#include "agentspace.h"
#include <cstdio>

using namespace std;

//------------------------- Block -----------------------------

void BaseBlock::setValidity (int ms) {
    if (ms == 0) {
        validity.tv_sec = 0;
        validity.tv_nsec = 0;
    }
    else {
        timespec now;
        Timer::get_time(&now);
        Timer::inc_time(&now,ms);
        validity = now;
    }
}

bool BaseBlock::isValid() {
    timespec now;
    if (erased) return FALSE;
    if (validity.tv_sec == 0 && validity.tv_nsec == 0) return TRUE;
    Timer::get_time(&now);
    return Timer::cmp_time(&now,&validity);
}

//------------------------- Space -----------------------------

Space* Space::space = NULL;

bool Space::wildcard (const string& pattern, const string& text) {
    const char *pat = pattern.c_str();
    const char *str = text.c_str();
    const char *retpat = NULL;
    const char *retstr = NULL;
    for (;*str;) {
        if (*pat == '?') {
            pat++; str++;
        }
        else if (*pat == '*') {
            while (*(pat+1) == '*') pat++;
            if (*(pat+1) == *str || *(pat+1) == '?') {
                retpat = pat;
                retstr = str+1;
                pat++;
            }
            else str++;
        }
        else if (*pat != *str) {
            if (retstr != NULL) {
                pat = retpat;
                str = retstr;
            }
            else return FALSE;
        }
        else {
            pat++; str++;
        }
    }
    while (*pat == '*') pat++;
    if (*pat) return FALSE;
    return TRUE;
}

bool Space::is_wildcard (const string& pattern) {
    const char *pat = pattern.c_str();
    for (; *pat; pat++)
        if (*pat == '*' || *pat == '?') return TRUE;
    return FALSE;
}

void Space::run(void *args) {
    Monitor monitor;
    Proxy *proxy = new Proxy(monitor);
    // add proxy to agent resources
    Timer *timer = new Timer(*proxy);
    // add timer to agent resouces
    timer->start(20,20,20);
    for (;;) {
        //cout << "\t Space is initialized !!!" << endl;
        Proxy::receive(monitor);
        sync.lock();
        map<string, BaseBlock *>::iterator tb;
        for (tb = table.begin(); tb != table.end(); tb++)
            if ((*tb).second != NULL){
                if (!(*tb).second->erased)
                    if (!(*tb).second->isValid()) {
                        BaseBlock *ret = (*tb).second;
                        ret->erased = TRUE;
                        set<Proxy *>::iterator pr;
                        for (pr=ret->triggers.begin(); pr!=ret->triggers.end(); pr++) {
                            if ((*pr)->is(TRIGGER_MATCHING))
                                matching((*pr)->pid(),ret);
                            (*pr)->trigger();
                        }
                    }
            }
        sync.unlock();
    }
}

void Space::matching (int pid, BaseBlock *bb) {
    set<BaseBlock *> *blocks = matching_blocks[pid];
    if (blocks == NULL) matching_blocks[pid] = blocks = new set<BaseBlock *>;
    blocks->insert(bb);
}

//------------------------- Agent -----------------------------

Agent::Agent(string args) : Thread(NULL) {
    //printf("Agent %d created\n", tid());
    _args = args;
    killed = FALSE;
}

Agent::~Agent() {
    // delete agent resources
}

unsigned int Agent::timer_attach (int start_ms, int period_ms, int round_ms) {
    Proxy *proxy = new Proxy(monitor);
    // add proxy to agent resources
    Timer *timer = new Timer(*proxy);
    // add timer to agent resouces
    timer->start(start_ms,period_ms,round_ms);
    return proxy->pid();
}

void Agent::run (void *args) {
    init(_args);
//    FILE_LOG(logDEBUG) << "*** Agent " << tid() << " initialized and running";
    for (;;) {
        if (killed) break;
        Proxy *proxy = Proxy::receive(monitor);
        sense_select_act(proxy->pid());
    }
//    FILE_LOG(logDEBUG) << "End of Agent " << tid();
}


void Agent::kill() {
    killed = TRUE;
}


