#ifndef agentspace_agentspace_INCLUDED
#define agentspace_agentspace_INCLUDED

#ifdef __GNUC__
#define TRUE 1
#define FALSE 0
#endif

#include "timer.h"
#include <map>
#include <set>
#include <list>

using namespace std;

// ---------------------- Space -----------------------
// unfortunately some compilers are not able to treat
// templates defined in more files (typically *.h and *.cpp)
// Therefore Space must be defined here

#define TRIGGER_NORMAL 1
#define TRIGGER_MATCHING 2

class BaseBlock {
    public:
        string name;
        struct timespec validity;
        float priority;
        bool erased;
        set<Proxy *> triggers;
        void setValidity (int ms);
        bool isValid();
};

template <typename T>
class Block : public BaseBlock {
    public:
        T value;
        Block (string& _name, T value) : value(value) {
            //cout << "konstruktor bloku\n ";
            priority = 1.0;
            erased = TRUE;
            name.assign(_name);
        };
};

class Space : public Thread {
    private:
        map<string, BaseBlock *> table;
        list< pair<string, Proxy *> > triggers;
        map<int, set<BaseBlock *> *> matching_blocks;
        static Space *space;
        Synchronize sync;
        Space() { };
        ~Space() { delete space;};
        void run(void *args);
        void matching (int pid, BaseBlock *bb);
    public:
        static Space *getSpace() {
            if (space == NULL) {
                space = new Space();
//                FILE_LOG(logDEBUG)  << "*** Space initialized" << endl;
            }
//            cout << "*** Space initialized" << endl;
            return space;
        }
        static bool wildcard (const string& mask, const string& text);
        static bool is_wildcard (const string& mask);
        template <typename T> inline void write (string name,T& value, int validity_ms=0, float priority=1.0) {
            sync.lock();
            BaseBlock *ret = table[name];
            if (ret == NULL) {
                Block<T>* rett = new Block<T>(name,value);
                table[name] = rett;
                ret = rett;
                list< pair<string, Proxy *> >::iterator tg;
                for (tg=triggers.begin(); tg!=triggers.end();) {
                    if (wildcard((*tg).first,name)) {
                        ret->triggers.insert((*tg).second);
                        if (!is_wildcard((*tg).first)) tg = triggers.erase(tg);
                        else tg++;
                    }
                }
            }
            else if (!ret->isValid() || priority >= ret->priority) {
                Block<T> *rett = static_cast<Block<T> *>(ret);
                rett->value = value;
            }
            ret->erased = FALSE;
            ret->setValidity(validity_ms);
            set<Proxy *>::iterator pr;
            for (pr=ret->triggers.begin(); pr!=ret->triggers.end(); pr++) {
                if ((*pr)->is(TRIGGER_MATCHING)) matching ((*pr)->pid(),ret);
                (*pr)->trigger();
            }
            sync.unlock();
        }
        template <typename T> inline const T& read (string name, T& deflt) {
            sync.lock();
            BaseBlock *ret = table[name];
            if (ret == NULL || !ret->isValid()) {
                sync.unlock();
                return deflt;
            }
            else {
                const Block<T> *rett = static_cast<const Block<T> *>(ret);
                sync.unlock();
                return rett->value;
            }

        }
        void trigger_attach (string mask,Proxy *proxy,int trigger_type) {
            bool found = FALSE;
            sync.lock();
            proxy->set(trigger_type);
            map<string, BaseBlock *>::iterator tb;
            for (tb = table.begin(); tb != table.end(); tb++)
                if (wildcard(mask,(*tb).first)) {
                    (*tb).second->triggers.insert(proxy);
                    found = TRUE;
                }
            if (is_wildcard(mask) || !found) {
                triggers.push_back(make_pair(mask,proxy));
            }
            sync.unlock();
        }
        // void trigger_detach (string mask, Proxy *proxy);
        set<BaseBlock *>::iterator read_begin (int pid) {
            set<BaseBlock *> *blocks = matching_blocks[pid];
            if (blocks == NULL) matching_blocks[pid] = blocks = new set<BaseBlock *>;
            return blocks->begin();
        }
        set<BaseBlock *>::iterator read_end (int pid) {
            set<BaseBlock *> *blocks = matching_blocks[pid];
            if (blocks == NULL) matching_blocks[pid] = blocks = new set<BaseBlock *>;
            return blocks->end();
        }
};

// ------------------- Space iterator -----------------

template <typename T>
class SpaceIterator {
    private:
        set<BaseBlock *>::iterator bi;
    public:
        void operator++() {
            bi++;
        }
        void operator++(int unused) {
            bi++;
        }
        Block<T> *operator*() {
            return static_cast<Block<T> *>(*bi);
        }
        Block<T> *operator->() {
            return static_cast<Block<T> *>(*bi);
        }
        void operator=(const set<BaseBlock *>::iterator& rhs) {
            bi = rhs;
        }
        bool operator!=(const set<BaseBlock *>::iterator& rhs) {
            return (bi != rhs);
        }
};

// ---------------------- Agent -----------------------

class Agent : public Thread {
    private:
        string _args;
        Monitor monitor;
        bool killed;
        void run(void *args);
    protected:
        Agent(string args);
        ~Agent();
        virtual void init (string args) = 0;
        virtual void sense_select_act (int pid) = 0;
        unsigned int timer_attach(int start_ms, int period_ms, int round_ms=1000);
        unsigned int trigger_attach(string block_mask,int trigger_type=TRIGGER_NORMAL) {
            Proxy *proxy = new Proxy(monitor);
            // add proxy to agent resources
            Space::getSpace()->trigger_attach(block_mask,proxy,trigger_type);
            // add trigger to agent resources
            return proxy->pid();
        };
        template <typename T> inline const T& space_read(string name,T& deflt) {
            return Space::getSpace()->read<T>(name,deflt);
        }
        template <typename T> inline void space_write(string name,T& value,int validity_ms=0,float priority=1.0) {
            Space::getSpace()->write<T>(name,value,validity_ms,priority);
        }
        set<BaseBlock *>::iterator space_read_begin (int pid) {
            return Space::getSpace()->read_begin(pid);
        }
        set<BaseBlock *>::iterator space_read_end (int pid) {
            return Space::getSpace()->read_end(pid);
        }
    public:
        void kill();
};

// --------- global function for Space maniplation -----------
#define Space_write(...) Space::getSpace()->write(__VA_ARGS__)
#define Space_read(...) Space::getSpace()->read(__VA_ARGS__)

// At the moment AgentSpace does not support arrays and constants



#endif
