#include "thread.h"

//-----------------------Thread-----------------------------

#ifdef __MINGW32__
#include <windows.h>
void delay (int ms) {
    Sleep(ms);
}
#elif __GNUC__
void delay(int ms){
    sleep(ms);
}
#endif

unsigned int Thread::nid = 0;

Thread::Thread(void *args) : started(false) {
    _nid = ++nid;
    start(args);
    //cout << "Bingo! Thread has been constructed" << endl;
}

Thread::~Thread() {
}

unsigned int Thread::tid() const {
    return _nid;
}

int Thread::start(void *args) {
    int ret;
    if (!started) {
        started = true;
        this->args = args;
        if ((ret = pthread_create(&_id, NULL, &Thread::exec, this)) != 0) return -1;
    }
    //cout << "Thread " << tid() << " constructed" << endl;
    return 0;
}

void Thread::join() {
    pthread_join(_id, NULL);
}

void *Thread::exec(void *thr) {
    Thread *thread = reinterpret_cast<Thread *> (thr);
    thread->run(thread->args);
    return NULL;
}

void Thread::delay (int ms) {
    ::delay(ms);
}

// --------------------Synchronize ------------------------

Synchronize::Synchronize() {
    pthread_mutex_init(&mutex, NULL);
}

Synchronize::~Synchronize() {
    pthread_mutex_destroy(&mutex);
}

void Synchronize::lock() {
    pthread_mutex_lock(&mutex);
}

void Synchronize::unlock() {
    pthread_mutex_unlock(&mutex);
}

// --------------------Monitor ------------------------

Monitor::Monitor() {
    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);
    last_msg = NULL;
}

Monitor::~Monitor() {
    pthread_cond_destroy(&cond);
    pthread_mutex_destroy(&mutex);
}

void *Monitor::wait() {
    //cout << "waitovanie na monitore\n";
    last_msg = NULL;
    pthread_mutex_lock(&mutex);
    pthread_cond_wait(&cond,&mutex);
    //cout << "tutok" << endl;
    pthread_mutex_unlock(&mutex);
    //cout << "tutok2" << endl;
    return last_msg;
}

void *Monitor::wait(struct timespec *tim) {
    last_msg = NULL;
    pthread_mutex_lock(&mutex);
    pthread_cond_timedwait(&cond,&mutex,tim);
    pthread_mutex_unlock(&mutex);
    return last_msg;
}

void Monitor::notify(void *msg) {
    last_msg = msg;
    pthread_mutex_lock(&mutex);
    pthread_cond_signal(&cond);
    pthread_mutex_unlock(&mutex);
}

//---------------------------Proxy--------------------------

unsigned int Proxy::nid = 0;

Proxy::Proxy(Monitor& monitor) {
    _nid = ++nid;
    _monitor = &monitor;
    status = 0;
}

Proxy::~Proxy() {
}

unsigned int Proxy::pid() const {
    return _nid;
}

void Proxy::trigger() {
    _monitor->notify(this);
}

Proxy* Proxy::receive (Monitor& monitor) {
    return reinterpret_cast<Proxy*>(monitor.wait());
}

bool Proxy::operator== (const Proxy& rhs) const {
    return (_nid == rhs._nid);
}

bool Proxy::is (int status_mask) {
    return ((status & status_mask) != 0);
}

void Proxy::set (int status_mask) {
    status = status | status_mask;
}

//-------------------------------------------------------

